# coding: utf-8

"""
    Optimeering

"""  # noqa: E501
from typing import Any, Dict, List, Optional, Tuple, Union

from optimeering.api_client import OptimeeringClient, RequestSerialized
from optimeering.logger import log_function_timing, suggest_series_id_optimization
from optimeering.models.predictions_data_list import PredictionsDataList
from optimeering.models.predictions_series_list import PredictionsSeriesList
from optimeering.models.predictions_version_list import PredictionsVersionList
from optimeering.models.versioned_series import VersionedSeries
from optimeering.silent_type_cast import silent_type_cast
from pydantic import Field, StrictBool, StrictFloat, StrictInt, StrictStr, validate_call
from typing_extensions import Annotated


class PredictionsApi:
    """
    Collection of methods to interact with PredictionsApi
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = OptimeeringClient.get_default()
        self.api_client = api_client

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def list_parameters(
        self,
        param: Optional[StrictStr],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> object:
        """Parameter Route

        Allowed values for each parameter used in filters on the **predictions** routes

        :param param: (required)
        :type param: str
        :return: Returns the result object.
        :rtype: object

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Post data point - replace ... with correct dataformat documented above
                >>> response = client.predictions_api.list_parameters(...)

        """  # noqa: E501

        _param = self._list_parameters_serialize(
            param=param,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        return response

    def _list_parameters_serialize(
        self,
        param,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if param is not None:
            _path_params["param"] = param
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/api/predictions/parameters/{param}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def list_series(
        self,
        id: Annotated[
            Optional[List[StrictInt]],
            Field(description="ID of the series to retrieve. If not specified, will return all series."),
        ] = None,
        product: Annotated[
            Optional[List[StrictStr]],
            Field(
                description="The product for which series should be retrieved. If not specified, will return series for all products."
            ),
        ] = None,
        unit_type: Annotated[
            Optional[List[StrictStr]],
            Field(description="Unit type. If not specified, will return series for all unit types."),
        ] = None,
        statistic: Annotated[
            Optional[List[StrictStr]],
            Field(description="Statistic type. If not specified, will return series for all statistic types."),
        ] = None,
        area: Annotated[
            Optional[List[StrictStr]],
            Field(description="The name of the area. If not specified, will return all areas."),
        ] = None,
        resolution: Annotated[
            Optional[List[StrictStr]],
            Field(description="Resolution of the series. If not specified, will return series for all resolutions."),
        ] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> PredictionsSeriesList:
        """List Series


        Returns prediction series.

        To list versions of a series, use the :any:`list_version` method.


        :param id: ID of the series to retrieve. If not specified, will return all series.
        :type id: List[StrictInt]
        :param product: The product for which series should be retrieved. If not specified, will return series for all products.
        :type product: List[StrictStr]
        :param unit_type: Unit type. If not specified, will return series for all unit types.
        :type unit_type: List[StrictStr]
        :param statistic: Statistic type. If not specified, will return series for all statistic types.
        :type statistic: List[StrictStr]
        :param area: The name of the area. If not specified, will return all areas.
        :type area: List[StrictStr]
        :param resolution: Resolution of the series. If not specified, will return series for all resolutions.
        :type resolution: List[StrictStr]
        :return: Returns the result object.
        :rtype: PredictionsSeriesList

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Get filtered data point - replace ... with appropriate filters documented above
                >>> response = client.predictions_api.list_series(...)

        """  # noqa: E501

        _param = self._list_series_serialize(
            id=id,
            product=product,
            unit_type=unit_type,
            statistic=statistic,
            area=area,
            resolution=resolution,
            limit=None,
            offset=None,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "PredictionsSeriesList",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        paginated_response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        next_page = paginated_response.next_page

        if next_page is None:
            return paginated_response

        # resolve extendable attribute
        attribute_to_extend_selected: Optional[str] = None
        if "series_id" in set(paginated_response.items[0].model_fields):
            extendable_attribute = {"datapoints", "events"}
            attribute_to_extend = extendable_attribute.intersection(set(paginated_response.items[0].model_fields))
            if len(attribute_to_extend) == 1:
                attribute_to_extend_selected = list(attribute_to_extend)[0]
            else:
                raise AttributeError("Failed to resolve extendable attribute.")

        while next_page is not None:
            next_pagination = self.api_client.call_api(
                method=_param[0],
                url=next_page,
                header_params=_param[2],
                body=_param[3],
                post_params=_param[4],
                _request_timeout=_request_timeout,
            )
            next_pagination.read()
            next_pagination = self.api_client.response_deserialize(
                response_data=next_pagination,
                response_types_map=_response_types_map,
            ).data
            next_page = next_pagination.next_page

            if attribute_to_extend_selected:
                if next_pagination.items[0].series_id == paginated_response.items[-1].series_id:
                    # if series id is same, extract the first item and extend to last item
                    continued_data: List = getattr(next_pagination.items.pop(0), attribute_to_extend_selected)
                    previous_data: List = getattr(paginated_response.items[-1], attribute_to_extend_selected)
                    previous_data.extend(continued_data)
            paginated_response.items.extend(next_pagination.items)

        return paginated_response

    def _list_series_serialize(
        self,
        id,
        product,
        unit_type,
        statistic,
        area,
        resolution,
        limit,
        offset,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if id is not None:
            id = ",".join(str(i) for i in id)
            _query_params.append(("id", id))

        if product is not None:
            product = ",".join(str(i) for i in product)
            _query_params.append(("product", product))

        if unit_type is not None:
            unit_type = ",".join(str(i) for i in unit_type)
            _query_params.append(("unit_type", unit_type))

        if statistic is not None:
            statistic = ",".join(str(i) for i in statistic)
            _query_params.append(("statistic", statistic))

        if area is not None:
            area = ",".join(str(i) for i in area)
            _query_params.append(("area", area))

        if resolution is not None:
            resolution = ",".join(str(i) for i in resolution)
            _query_params.append(("resolution", resolution))

        if limit is not None:
            _query_params.append(("limit", limit))

        if offset is not None:
            _query_params.append(("offset", offset))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/api/predictions/series",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def list_version(
        self,
        id: Annotated[
            Optional[List[StrictInt]],
            Field(description="ID of the series to retrieve. If not specified, will return all series."),
        ] = None,
        product: Annotated[
            Optional[List[StrictStr]],
            Field(
                description="The product for which series should be retrieved. If not specified, will return series for all products."
            ),
        ] = None,
        unit_type: Annotated[
            Optional[List[StrictStr]],
            Field(description="Unit type. If not specified, will return series for all unit types."),
        ] = None,
        statistic: Annotated[
            Optional[List[StrictStr]],
            Field(description="Statistic type. If not specified, will return series for all statistic types."),
        ] = None,
        area: Annotated[
            Optional[List[StrictStr]],
            Field(description="The name of the area. If not specified, will return all areas."),
        ] = None,
        resolution: Annotated[
            Optional[List[StrictStr]],
            Field(description="Resolution of the series. If not specified, will return series for all resolutions."),
        ] = None,
        limit: Annotated[
            Optional[Annotated[int, Field(le=100000, strict=True, ge=1)]], Field(description="Number of items per page")
        ] = None,
        offset: Annotated[Optional[Annotated[int, Field(strict=True, ge=0)]], Field(description="Page offset")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> PredictionsVersionList:
        """List Series Version


        Returns prediction series and their versions.



        :param id: ID of the series to retrieve. If not specified, will return all series.
        :type id: str
        :param product: The product for which series should be retrieved. If not specified, will return series for all products.
        :type product: str
        :param unit_type: Unit type. If not specified, will return series for all unit types.
        :type unit_type: str
        :param statistic: Statistic type. If not specified, will return series for all statistic types.
        :type statistic: str
        :param area: The name of the area. If not specified, will return all areas.
        :type area: str
        :param resolution: Resolution of the series. If not specified, will return series for all resolutions.
        :type resolution: str
        :param limit: Number of items per page
        :type limit: int
        :param offset: Page offset
        :type offset: int
        :return: Returns the result object.
        :rtype: PredictionsVersionList

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Post data point - replace ... with correct dataformat documented above
                >>> response = client.predictions_api.list_version(...)

        """  # noqa: E501

        _param = self._list_version_serialize(
            id=id,
            product=product,
            unit_type=unit_type,
            statistic=statistic,
            area=area,
            resolution=resolution,
            limit=limit,
            offset=offset,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "PredictionsVersionList",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        return response

    def _list_version_serialize(
        self,
        id,
        product,
        unit_type,
        statistic,
        area,
        resolution,
        limit,
        offset,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if id is not None:
            id = ",".join(str(i) for i in id)
            _query_params.append(("id", id))

        if product is not None:
            product = ",".join(str(i) for i in product)
            _query_params.append(("product", product))

        if unit_type is not None:
            unit_type = ",".join(str(i) for i in unit_type)
            _query_params.append(("unit_type", unit_type))

        if statistic is not None:
            statistic = ",".join(str(i) for i in statistic)
            _query_params.append(("statistic", statistic))

        if area is not None:
            area = ",".join(str(i) for i in area)
            _query_params.append(("area", area))

        if resolution is not None:
            resolution = ",".join(str(i) for i in resolution)
            _query_params.append(("resolution", resolution))

        if limit is not None:
            _query_params.append(("limit", limit))

        if offset is not None:
            _query_params.append(("offset", offset))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/api/predictions/series_version",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def retrieve(
        self,
        series_id: Annotated[
            Optional[List[StrictInt]],
            Field(description="Series ID to filter. If not specified, will return an exception."),
        ] = None,
        start: Annotated[
            Optional[Any],
            Field(
                description="The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)"
            ),
        ] = None,
        end: Annotated[
            Optional[Any],
            Field(
                description="The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)"
            ),
        ] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> PredictionsDataList:
        """Retrieve


        Returns predictions.

        If multiple versions of a prediction exist for a given series, the highest version is returned.

        To get predictions for a particular version, use the :any:`retrieve_versioned` method.


        :param series_id: Series ID to filter. If not specified, will return an exception.
        :type series_id: List[StrictInt]
        :param start: The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        :type start: Start
        :param end: The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        :type end: End
        :return: Returns the result object.
        :rtype: PredictionsDataList

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Get filtered data point - replace ... with appropriate filters documented above
                >>> response = client.predictions_api.retrieve(...)

        """  # noqa: E501

        _param = self._retrieve_serialize(
            series_id=series_id,
            start=start,
            end=end,
            limit=None,
            offset=None,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "PredictionsDataList",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        paginated_response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        next_page = paginated_response.next_page

        if next_page is None:
            return paginated_response

        # resolve extendable attribute
        attribute_to_extend_selected: Optional[str] = None
        if "series_id" in set(paginated_response.items[0].model_fields):
            extendable_attribute = {"datapoints", "events"}
            attribute_to_extend = extendable_attribute.intersection(set(paginated_response.items[0].model_fields))
            if len(attribute_to_extend) == 1:
                attribute_to_extend_selected = list(attribute_to_extend)[0]
            else:
                raise AttributeError("Failed to resolve extendable attribute.")

        while next_page is not None:
            next_pagination = self.api_client.call_api(
                method=_param[0],
                url=next_page,
                header_params=_param[2],
                body=_param[3],
                post_params=_param[4],
                _request_timeout=_request_timeout,
            )
            next_pagination.read()
            next_pagination = self.api_client.response_deserialize(
                response_data=next_pagination,
                response_types_map=_response_types_map,
            ).data
            next_page = next_pagination.next_page

            if attribute_to_extend_selected:
                if next_pagination.items[0].series_id == paginated_response.items[-1].series_id:
                    # if series id is same, extract the first item and extend to last item
                    continued_data: List = getattr(next_pagination.items.pop(0), attribute_to_extend_selected)
                    previous_data: List = getattr(paginated_response.items[-1], attribute_to_extend_selected)
                    previous_data.extend(continued_data)
            paginated_response.items.extend(next_pagination.items)

        return paginated_response

    def _retrieve_serialize(
        self,
        series_id,
        start,
        end,
        limit,
        offset,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if series_id is not None:
            series_id = ",".join(str(i) for i in series_id)
            _query_params.append(("series_id", series_id))

        if start is not None:
            _query_params.append(("start", start))

        if end is not None:
            _query_params.append(("end", end))

        if limit is not None:
            _query_params.append(("limit", limit))

        if offset is not None:
            _query_params.append(("offset", offset))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/api/predictions",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def retrieve_latest(
        self,
        max_event_time: Annotated[
            Optional[Any],
            Field(
                description="If specified, will only return the latest prediction available at the specified time. If not specified, no filters are applied. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)"
            ),
        ] = None,
        series_id: Annotated[
            Optional[List[StrictInt]],
            Field(description="Series ID to filter. If not specified, will return an exception."),
        ] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> PredictionsDataList:
        """Retrieve Latest


        Returns predictions with the most recent ``event_time``.

        If multiple versions of a prediction exist for a given series, the highest version is returned.

        To get predictions for a particular version, use the :any:`retrieve_versioned`  method.


        :param max_event_time: If specified, will only return the latest prediction available at the specified time. If not specified, no filters are applied. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        :type max_event_time: MaxEventTime
        :param series_id: Series ID to filter. If not specified, will return an exception.
        :type series_id: List[StrictInt]
        :return: Returns the result object.
        :rtype: PredictionsDataList

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Get filtered data point - replace ... with appropriate filters documented above
                >>> response = client.predictions_api.retrieve_latest(...)

        """  # noqa: E501

        _param = self._retrieve_latest_serialize(
            max_event_time=max_event_time,
            series_id=series_id,
            limit=None,
            offset=None,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "PredictionsDataList",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        paginated_response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        next_page = paginated_response.next_page

        if next_page is None:
            return paginated_response

        # resolve extendable attribute
        attribute_to_extend_selected: Optional[str] = None
        if "series_id" in set(paginated_response.items[0].model_fields):
            extendable_attribute = {"datapoints", "events"}
            attribute_to_extend = extendable_attribute.intersection(set(paginated_response.items[0].model_fields))
            if len(attribute_to_extend) == 1:
                attribute_to_extend_selected = list(attribute_to_extend)[0]
            else:
                raise AttributeError("Failed to resolve extendable attribute.")

        while next_page is not None:
            next_pagination = self.api_client.call_api(
                method=_param[0],
                url=next_page,
                header_params=_param[2],
                body=_param[3],
                post_params=_param[4],
                _request_timeout=_request_timeout,
            )
            next_pagination.read()
            next_pagination = self.api_client.response_deserialize(
                response_data=next_pagination,
                response_types_map=_response_types_map,
            ).data
            next_page = next_pagination.next_page

            if attribute_to_extend_selected:
                if next_pagination.items[0].series_id == paginated_response.items[-1].series_id:
                    # if series id is same, extract the first item and extend to last item
                    continued_data: List = getattr(next_pagination.items.pop(0), attribute_to_extend_selected)
                    previous_data: List = getattr(paginated_response.items[-1], attribute_to_extend_selected)
                    previous_data.extend(continued_data)
            paginated_response.items.extend(next_pagination.items)

        return paginated_response

    def _retrieve_latest_serialize(
        self,
        max_event_time,
        series_id,
        limit,
        offset,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if max_event_time is not None:
            _query_params.append(("max_event_time", max_event_time))

        if series_id is not None:
            series_id = ",".join(str(i) for i in series_id)
            _query_params.append(("series_id", series_id))

        if limit is not None:
            _query_params.append(("limit", limit))

        if offset is not None:
            _query_params.append(("offset", offset))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/api/predictions/latest",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @silent_type_cast
    @suggest_series_id_optimization
    def retrieve_versioned(
        self,
        include_simulated: Annotated[
            Optional[StrictBool], Field(description="If false, filters out simulated prediction from response.")
        ] = None,
        start: Annotated[
            Optional[Any],
            Field(
                description="The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)"
            ),
        ] = None,
        end: Annotated[
            Optional[Any],
            Field(
                description="The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)"
            ),
        ] = None,
        versioned_series: Optional[List[VersionedSeries] | PredictionsVersionList] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> PredictionsDataList:
        """Retrive Versioned


        Returns versioned predictions.


        Use the :any:`list_version` method to get the available versions for each prediction series.

        Can be used to retrieve both versioned and simulated data. For an explanation on versioned and simulated data see `Prediction Versioning <https://docs.optimeering.com/getting-started/prediction-versioning/>`_


        :param include_simulated: If false, filters out simulated prediction from response.
        :type include_simulated: bool
        :param start: The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        :type start: Start
        :param end: The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        :type end: End
        :param versioned_series:
        :type versioned_series: Optional[List[VersionedSeries] | PredictionsVersionList]
        :return: Returns the result object.
        :rtype: PredictionsDataList

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Get filtered data point - replace ... with appropriate filters documented above
                >>> response = client.predictions_api.retrieve_versioned(...)

        """  # noqa: E501

        _param = self._retrieve_versioned_serialize(
            include_simulated=include_simulated,
            start=start,
            end=end,
            limit=None,
            offset=None,
            versioned_series=versioned_series,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "PredictionsDataList",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        paginated_response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        next_page = paginated_response.next_page

        if next_page is None:
            return paginated_response

        # resolve extendable attribute
        attribute_to_extend_selected: Optional[str] = None
        if "series_id" in set(paginated_response.items[0].model_fields):
            extendable_attribute = {"datapoints", "events"}
            attribute_to_extend = extendable_attribute.intersection(set(paginated_response.items[0].model_fields))
            if len(attribute_to_extend) == 1:
                attribute_to_extend_selected = list(attribute_to_extend)[0]
            else:
                raise AttributeError("Failed to resolve extendable attribute.")

        while next_page is not None:
            next_pagination = self.api_client.call_api(
                method=_param[0],
                url=next_page,
                header_params=_param[2],
                body=_param[3],
                post_params=_param[4],
                _request_timeout=_request_timeout,
            )
            next_pagination.read()
            next_pagination = self.api_client.response_deserialize(
                response_data=next_pagination,
                response_types_map=_response_types_map,
            ).data
            next_page = next_pagination.next_page

            if attribute_to_extend_selected:
                if next_pagination.items[0].series_id == paginated_response.items[-1].series_id:
                    # if series id is same, extract the first item and extend to last item
                    continued_data: List = getattr(next_pagination.items.pop(0), attribute_to_extend_selected)
                    previous_data: List = getattr(paginated_response.items[-1], attribute_to_extend_selected)
                    previous_data.extend(continued_data)
            paginated_response.items.extend(next_pagination.items)

        return paginated_response

    def _retrieve_versioned_serialize(
        self,
        include_simulated,
        start,
        end,
        limit,
        offset,
        versioned_series,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {
            "VersionedSeries": "",
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if include_simulated is not None:
            _query_params.append(("include_simulated", include_simulated))

        if start is not None:
            _query_params.append(("start", start))

        if end is not None:
            _query_params.append(("end", end))

        if limit is not None:
            _query_params.append(("limit", limit))

        if offset is not None:
            _query_params.append(("offset", offset))

        # process the header parameters
        # process the form parameters
        # process the body parameter
        if versioned_series is not None:
            _body_params = versioned_series

        return self.api_client.param_serialize(
            method="POST",
            resource_path="/api/predictions/versioned",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
import warnings
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Set

import orjson
from optimeering.extras import pd, pydantic_to_pandas, require_pandas
from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, model_validator


class AccessKeyPostResponse(BaseModel):
    """
    AccessKeyPostResponse

    :param apikey: API key
    :type apikey: str
    :param created_at: Time stamp at which key was created.
    :type created_at: datetime
    :param description: Description for the Access key.
    :type description: str
    :param expires_at: Duration after which key expires.
    :type expires_at: datetime
    :param id: ID of the access key
    :type id: int
    :param owner_id: Creator of the access key.
    :type owner_id: str
    """  # noqa: E501

    apikey: StrictStr = Field(description="API key")
    created_at: datetime = Field(description="Time stamp at which key was created.")
    description: StrictStr = Field(description="Description for the Access key.")
    expires_at: datetime = Field(description="Duration after which key expires.")
    id: StrictInt = Field(description="ID of the access key")
    owner_id: StrictStr = Field(description="Creator of the access key.")

    __properties: ClassVar[List[str]] = ["apikey", "created_at", "description", "expires_at", "id", "owner_id"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return orjson.dumps(self.to_dict()).decode()

    @classmethod
    def from_json(cls, json_str: str) -> Optional[AccessKeyPostResponse]:
        """Create an instance of AccessKeyPostResponse from a JSON string"""
        return cls.from_dict(orjson.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[AccessKeyPostResponse]:
        """Create an instance of AccessKeyPostResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "apikey": obj.get("apikey"),
                "created_at": obj.get("created_at"),
                "description": obj.get("description"),
                "expires_at": obj.get("expires_at"),
                "id": obj.get("id"),
                "owner_id": obj.get("owner_id"),
            }
        )
        return _obj

    @model_validator(mode="before")
    def validate_extra_fields(cls, values):
        if isinstance(values, Dict):
            values_private_removed = {k: v for k, v in values.items() if not k.startswith("_")}
        else:
            values_private_removed = values
        if len(values_private_removed) > 1:  # Check if there are extra fields
            if set(values_private_removed) - set(cls.model_fields):
                warnings.warn("Data mismatch, please update the SDK to the latest version")
        return values

    def __len__(self):
        return 1

    @require_pandas
    def to_pandas(
        self,
    ) -> "pd.DataFrame":  # type: ignore[name-defined]
        """
        Converts the object into a pandas dataframe.

        """
        return pydantic_to_pandas(
            self,
        )



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
import warnings
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Set

import optimeering
import orjson
from optimeering.extras import pd, pydantic_to_pandas, require_pandas
from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator, model_validator
from typing_extensions import Annotated


class PredictionsVersion(BaseModel):
    """
    A :any:`PredictionsVersion` is used for indexing a specific version for a series of :any:`PredictionsData`. For an explanation on versioned and simulated data see `Prediction Versioning <https://docs.optimeering.com/getting-started/prediction-versioning/>`_

    :param area: Areas to be filtered. E.g. NO1, NO2
    :type area: str
    :param created_at:
    :type created_at: datetime
    :param description:
    :type description: str
    :param id:
    :type id: int
    :param latest_event_time:
    :type latest_event_time: datetime
    :param product: Product name for the series
    :type product: str
    :param resolution: Resolution of the series.
    :type resolution: str
    :param simulation_event_time_end: The timestamp to which predictions is generated using simulation
    :type simulation_event_time_end: datetime
    :param simulation_event_time_start: The timestamp from which predictions is generated using simulation
    :type simulation_event_time_start: datetime
    :param statistic: Type of statistics.
    :type statistic: str
    :param unit: The unit for the series.
    :type unit: str
    :param unit_type: Unit type for the series
    :type unit_type: str
    :param version: Version of the model that generated the predictions
    :type version: str
    """  # noqa: E501

    area: StrictStr = Field(description="Areas to be filtered. E.g. NO1, NO2")
    created_at: datetime
    description: Optional[StrictStr] = None
    id: StrictInt
    latest_event_time: Optional[datetime] = None
    product: StrictStr = Field(description="Product name for the series")
    resolution: StrictStr = Field(description="Resolution of the series.")
    simulation_event_time_end: datetime = Field(
        description="The timestamp to which predictions is generated using simulation"
    )
    simulation_event_time_start: datetime = Field(
        description="The timestamp from which predictions is generated using simulation"
    )
    statistic: StrictStr = Field(description="Type of statistics.")
    unit: StrictStr = Field(description="The unit for the series.")
    unit_type: StrictStr = Field(description="Unit type for the series")
    version: Annotated[str, Field(strict=True)] = Field(
        description="Version of the model that generated the predictions"
    )
    _client: Any = None

    __properties: ClassVar[List[str]] = [
        "area",
        "created_at",
        "description",
        "id",
        "latest_event_time",
        "product",
        "resolution",
        "simulation_event_time_end",
        "simulation_event_time_start",
        "statistic",
        "unit",
        "unit_type",
        "version",
    ]

    @field_validator("version")
    def version_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"[0-9]+\.[0-9]+\.[0-9]+", value):
            raise ValueError(r"must validate the regular expression /[0-9]+\.[0-9]+\.[0-9]+/")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return orjson.dumps(self.to_dict()).decode()

    @classmethod
    def from_json(cls, json_str: str) -> Optional[PredictionsVersion]:
        """Create an instance of PredictionsVersion from a JSON string"""
        return cls.from_dict(orjson.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if description (nullable) is None
        # and model_fields_set contains the field
        if self.description is None and "description" in self.model_fields_set:
            _dict["description"] = None

        # set to None if latest_event_time (nullable) is None
        # and model_fields_set contains the field
        if self.latest_event_time is None and "latest_event_time" in self.model_fields_set:
            _dict["latest_event_time"] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[PredictionsVersion]:
        """Create an instance of PredictionsVersion from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "area": obj.get("area"),
                "created_at": obj.get("created_at"),
                "description": obj.get("description"),
                "id": obj.get("id"),
                "latest_event_time": obj.get("latest_event_time"),
                "product": obj.get("product"),
                "resolution": obj.get("resolution"),
                "simulation_event_time_end": obj.get("simulation_event_time_end"),
                "simulation_event_time_start": obj.get("simulation_event_time_start"),
                "statistic": obj.get("statistic"),
                "unit": obj.get("unit"),
                "unit_type": obj.get("unit_type"),
                "version": obj.get("version"),
            }
        )
        return _obj

    @model_validator(mode="before")
    def validate_extra_fields(cls, values):
        if isinstance(values, Dict):
            values_private_removed = {k: v for k, v in values.items() if not k.startswith("_")}
        else:
            values_private_removed = values
        if len(values_private_removed) > 1:  # Check if there are extra fields
            if set(values_private_removed) - set(cls.model_fields):
                warnings.warn("Data mismatch, please update the SDK to the latest version")
        return values

    def retrieve_versioned(
        self,
        include_simulated: Optional[bool] = None,
        start: Optional[datetime | StrictStr] = None,
        end: Optional[datetime | StrictStr] = None,
    ) -> optimeering.models.PredictionsDataList:
        """

        Returns versioned predictions.


        Use the :any:`list_version` method to get the available versions for each prediction series.

        Can be used to retrieve both versioned and simulated data. For an explanation on versioned and simulated data see `Prediction Versioning <https://docs.optimeering.com/getting-started/prediction-versioning/>`_


                :param include_simulated: If false, filters out simulated prediction from response.
                :param start: The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
                :param end: The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        """
        if self._client is None:
            raise AttributeError("Cannot call datapoints method on this instance. The client has not been setup.")

        return optimeering.PredictionsApi(api_client=self._client).retrieve_versioned(
            versioned_series=[self],
            include_simulated=include_simulated,
            start=start,
            end=end,
        )

    def __len__(self):
        return 1

    def convert_to_versioned_series(self) -> optimeering.VersionedSeries:
        """Convert to VersionedSeries"""

        return optimeering.VersionedSeries(
            series_id=self.id,
            version=self.version,
        )

    @require_pandas
    def to_pandas(
        self,
    ) -> "pd.DataFrame":  # type: ignore[name-defined]
        """
        Converts the object into a pandas dataframe.

        """
        return pydantic_to_pandas(
            self,
        )



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

import copy
import http.client as httplib
import logging
import multiprocessing
import sys
from logging import FileHandler
from typing import Optional, Union
from urllib.parse import urlparse

import urllib3

JSON_SCHEMA_VALIDATION_KEYWORDS = {
    "multipleOf",
    "maximum",
    "exclusiveMaximum",
    "minimum",
    "exclusiveMinimum",
    "maxLength",
    "minLength",
    "pattern",
    "maxItems",
    "minItems",
}


class Configuration:
    """This class is used to configure the :any:`OptimeeringClient`.

    :param host: API host url.
    :param api_auth_url: URL for Authentication. Defaults to `api://{host}/.default`
    :param retries:
        Number of retries for API requests.
        Also supports a urllib ``Retry`` object. `See the urlib docs <https://urllib3.readthedocs.io/en/stable/reference/urllib3.util.html#urllib3.util.Retry>`_
    :param api_key: API Key to use for Authentication.
    :param token: Token to use for Authentication. Only use this when you don't want to have the :any:`OptimeeringClient` retrieve a token.

    :Example:

    >>> from optimeering import Configuration, OptimeeringClient
    >>> configuration = Configuration(host="https://app.optimeering.com")

    :API Key Example:

    >>> from optimeering import Configuration, OptimeeringClient
    >>> configuration = Configuration(host="https://app.optimeering.com", api_key="your api key")

    :Token Example:

    >>> from optimeering import Configuration, OptimeeringClient
    >>> configuration = Configuration(host="https://app.optimeering.com", token="your token")
    """

    _default = None

    def __init__(
        self,
        host: Optional[str] = "https://app.optimeering.com",
        api_auth_url: Optional[str] = None,
        *,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        retries: Optional[Union[int, urllib3.Retry]] = None,
        debug: Optional[bool] = None,
    ) -> None:
        """Constructor"""
        self._base_path = "http://localhost" if host is None else host
        self.api_auth_url = (
            f"api://{urlparse(self._base_path).netloc}/.default" if api_auth_url is None else api_auth_url
        )

        # Check auth conflict
        auth_parameters = (api_key, token)
        if sum((auth_param is not None) for auth_param in auth_parameters) > 1:
            raise Exception("Only a single mode of authorization can be used.")

        self.api_key = api_key
        self.token = token

        self.logger = {}
        """Logging Settings
        """
        self.logger["package_logger"] = logging.getLogger("optimeering")
        self.logger["urllib3_logger"] = logging.getLogger("urllib3")
        self.logger_format = "%(asctime)s %(levelname)s %(message)s"
        """Log format
        """
        self.logger_stream_handler = None
        """Log stream handler
        """
        self.logger_file_handler: Optional[FileHandler] = None
        """Log file handler
        """
        self.logger_file = None
        """Debug file location
        """
        if debug is not None:
            self.debug = debug
        else:
            self.__debug = False
        """Debug switch
        """

        self.verify_ssl = True
        """SSL/TLS verification
           Set this to false to skip verifying SSL certificate when calling API
           from https server.
        """
        self.ssl_ca_cert = None
        """Set this to customize the certificate file to verify the peer.
        """
        self.cert_file = None
        """client certificate file
        """
        self.key_file = None
        """client key file
        """
        self.assert_hostname = None
        """Set this to True/False to enable/disable SSL hostname verification.
        """
        self.tls_server_name = None
        """SSL/TLS Server Name Indication (SNI)
           Set this to the SNI value expected by the server.
        """

        self.connection_pool_maxsize = multiprocessing.cpu_count() * 5
        """urllib3 connection pool's maximum number of connections saved
           per pool. urllib3 uses 1 connection as default value, but this is
           not the best value when you are making a lot of possibly parallel
           requests to the same host, which is often the case here.
           cpu_count * 5 is used as default value to increase performance.
        """

        self.proxy: Optional[str] = None
        """Proxy URL
        """
        self.proxy_headers = None
        """Proxy headers
        """
        self.retries = retries
        """Adding retries to override urllib3 default value 3
        """
        # Enable client side validation
        self.client_side_validation = True

        self.socket_options = None
        """Options to pass down to the underlying urllib3 socket
        """

        self.datetime_format = "%Y-%m-%dT%H:%M:%S.%f%z"
        """datetime format
        """

        self.date_format = "%Y-%m-%d"
        """date format
        """

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ("logger", "logger_file_handler"):
                setattr(result, k, copy.deepcopy(v, memo))
        # shallow copy of loggers
        result.logger = copy.copy(self.logger)
        # use setters to configure loggers
        result.logger_file = self.logger_file
        result.debug = self.debug
        return result

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

    @classmethod
    def set_default(cls, default):
        """Set default instance of configuration.

        It stores default configuration, which can be
        returned by get_default_copy method.

        :param default: object of Configuration
        """
        cls._default = default

    @classmethod
    def get_default_copy(cls):
        """Deprecated. Please use `get_default` instead.

        Deprecated. Please use `get_default` instead.

        :return: The configuration object.
        """
        return cls.get_default()

    @classmethod
    def get_default(cls):
        """Return the default configuration.

        This method returns newly created, based on default constructor,
        object of Configuration class or returns a copy of default
        configuration.

        :return: The configuration object.
        """
        if cls._default is None:
            cls._default = Configuration()
        return cls._default

    @property
    def logger_file(self):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :type: str
        """
        return self.__logger_file

    @logger_file.setter
    def logger_file(self, value):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        self.__logger_file = value
        if self.__logger_file:
            # If set logging file,
            # then add file handler and remove stream handler.
            self.logger_file_handler = logging.FileHandler(self.__logger_file)
            self.logger_file_handler.setFormatter(self.logger_formatter)
            for _, logger in self.logger.items():
                logger.addHandler(self.logger_file_handler)

    @property
    def debug(self):
        """Debug status"""
        return self.__debug

    @debug.setter
    def debug(self, value):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        self.__debug = value
        if self.__debug:
            # if debug status is True, turn on debug logging
            for _, logger in self.logger.items():
                logger.setLevel(logging.DEBUG)
            # turn on httplib debug
            httplib.HTTPConnection.debuglevel = 1
        else:
            # if debug status is False, turn off debug logging,
            # setting log level to default `logging.WARNING`
            for _, logger in self.logger.items():
                logger.setLevel(logging.WARNING)
            # turn off httplib debug
            httplib.HTTPConnection.debuglevel = 0

    @property
    def logger_format(self):
        """The logger format.
        The logger_formatter will be updated when sets logger_format.
        """
        return self.__logger_format

    @logger_format.setter
    def logger_format(self, value):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        self.__logger_format = value
        self.logger_formatter = logging.Formatter(self.__logger_format)

    def to_debug_report(self):
        """Gets the essential information for debugging.

        :return: The report for debugging.
        """
        return (
            "Python SDK Debug Report:\n"
            "OS: {env}\n"
            "Python Version: {pyversion}\n"
            "Version of the API: 0.1.0\n"
            "SDK Package Version: 0.0.2".format(env=sys.platform, pyversion=sys.version)
        )

    @property
    def host(self) -> str:
        return self._base_path



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
import warnings
from datetime import datetime
from typing import Any, ClassVar, Dict, Iterator, List, Optional, Set

import optimeering
import orjson
from optimeering.extras import pd, pydantic_to_pandas, require_pandas
from optimeering.models.predictions_version import PredictionsVersion
from pydantic import BaseModel, ConfigDict, Field, StrictStr, model_validator


class PredictionsVersionList(BaseModel):
    """
    A :any:`PredictionsVersionList` contains a collection of :any:`PredictionsVersion`.

    :param items:
    :type items: List[PredictionsVersion]
    :param next_page: The next page of results (if available).
    :type next_page: str
    """  # noqa: E501

    items: List[PredictionsVersion]
    next_page: Optional[StrictStr] = Field(default=None, description="The next page of results (if available).")
    _client: Any = None

    __properties: ClassVar[List[str]] = ["items", "next_page"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return orjson.dumps(self.to_dict()).decode()

    @classmethod
    def from_json(cls, json_str: str) -> Optional[PredictionsVersionList]:
        """Create an instance of PredictionsVersionList from a JSON string"""
        return cls.from_dict(orjson.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in items (list)
        _items = []
        if self.items:
            for _item_items in self.items:
                if _item_items:
                    _items.append(_item_items.to_dict())
            _dict["items"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[PredictionsVersionList]:
        """Create an instance of PredictionsVersionList from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "items": [PredictionsVersion.from_dict(_item) for _item in obj["items"]]
                if obj.get("items") is not None
                else None,
                "next_page": obj.get("next_page"),
            }
        )
        return _obj

    @model_validator(mode="before")
    def validate_extra_fields(cls, values):
        if isinstance(values, Dict):
            values_private_removed = {k: v for k, v in values.items() if not k.startswith("_")}
        else:
            values_private_removed = values
        if len(values_private_removed) > 1:  # Check if there are extra fields
            if set(values_private_removed) - set(cls.model_fields):
                warnings.warn("Data mismatch, please update the SDK to the latest version")
        return values

    def retrieve_versioned(
        self,
        include_simulated: Optional[bool] = None,
        start: Optional[datetime | StrictStr] = None,
        end: Optional[datetime | StrictStr] = None,
    ) -> optimeering.models.PredictionsDataList:
        """

        Returns versioned predictions.


        Use the :any:`list_version` method to get the available versions for each prediction series.

        Can be used to retrieve both versioned and simulated data. For an explanation on versioned and simulated data see `Prediction Versioning <https://docs.optimeering.com/getting-started/prediction-versioning/>`_


                :param include_simulated: If false, filters out simulated prediction from response.
                :param start: The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
                :param end: The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        """
        if self._client is None:
            raise AttributeError("Cannot call datapoints method on this instance. The client has not been setup.")

        return optimeering.PredictionsApi(api_client=self._client).retrieve_versioned(
            versioned_series=self,
            include_simulated=include_simulated,
            start=start,
            end=end,
        )

    def __len__(self):
        return sum(len(i) for i in self.items)

    def __iter__(self) -> Iterator[PredictionsVersion]:  # type: ignore[override]
        """Iteration method for generated models"""
        self.__iter_index = 0
        return self

    def __next__(self) -> PredictionsVersion:
        if not self.items:
            raise StopIteration
        try:
            _return = self.items[self.__iter_index]
        except IndexError:
            del self.__iter_index
            raise StopIteration
        else:
            self.__iter_index += 1
            return _return

    def filter(
        self,
        area: Optional[List[str]] = None,
        id: Optional[List[int]] = None,
        product: Optional[List[str]] = None,
        resolution: Optional[List[str]] = None,
        statistic: Optional[List[str]] = None,
        unit: Optional[List[str]] = None,
        unit_type: Optional[List[str]] = None,
        version: Optional[List[str]] = None,
    ) -> PredictionsVersionList:
        """Filters items"""
        properties = [
            "area",
            "id",
            "product",
            "resolution",
            "statistic",
            "unit",
            "unit_type",
            "version",
        ]
        _locals = locals()
        compare_dict = {i: _locals[i] for i in properties if _locals[i] is not None}

        obj_copy = self.model_copy()
        obj_copy.items = [
            item
            for item in obj_copy
            if all(
                getattr(item, property_name) in filter_values for property_name, filter_values in compare_dict.items()
            )
        ]
        return obj_copy

    def convert_to_versioned_series(self) -> List:
        """Converts all items"""
        return [i.convert_to_versioned_series() for i in self.items]

    @require_pandas
    def to_pandas(
        self,
    ) -> "pd.DataFrame":  # type: ignore[name-defined]
        """
        Converts the object into a pandas dataframe.

        """
        return pydantic_to_pandas(
            self,
        )



import inspect
import re
import typing
from functools import wraps


def silent_type_cast(fn):
    """
    Decorator that attempts to silently type cast args and kwargs (if supported)
    """
    fn_annotations = inspect.signature(fn)
    parameter_conversion = {}
    for param_name, annotation in fn_annotations.parameters.items():
        base_type = typing.get_args(annotation.annotation)
        if base_type:  # continue only if method has parameters
            while True:
                if isinstance(base_type, tuple):
                    # if iterable, take the first one
                    base_type = base_type[0]
                elif typing.get_origin(base_type) in (list, typing.Union):
                    # if iterable type, take the first one
                    base_type = typing.get_args(base_type)
                    if not base_type:
                        break
                    base_type = base_type[0]
                elif base_type is typing.Any:
                    # no need of polymorphism if any type
                    break
                elif base_type.__module__ == "builtins":
                    # if one of the base python object, polymorphism is not supported
                    break
                else:
                    class_name = base_type.__name__
                    function_name = f"convert_to_{re.sub(r'(?<!^)(?=[A-Z])', '_', class_name).lower()}"
                    parameter_conversion[param_name] = function_name
                    break

    @wraps(fn)
    def decorator(*args, **kwargs):
        new_args = []
        for arg, _param_name in zip(args, fn_annotations.parameters.keys()):
            if (
                arg is None
                or _param_name not in parameter_conversion
                or not hasattr(arg, parameter_conversion[_param_name])
            ):
                new_args.append(arg)
            else:
                new_args.append(getattr(arg, parameter_conversion[_param_name])())
        new_args_tuple = tuple(new_args)
        kwargs = {
            k: getattr(v, parameter_conversion[k])()
            if v  # not null
            and k in parameter_conversion  # parameter allows conversion
            and hasattr(v, parameter_conversion[k])  # conversion fn exists
            else v
            for k, v in kwargs.items()
        }
        return fn(*new_args_tuple, **kwargs)

    return decorator



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Set, Union

import orjson
from pydantic import BaseModel, ValidationError, field_validator

START_ANY_OF_SCHEMAS = ["datetime", "str"]


class Start(BaseModel):
    """
    The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
    """

    # data type: str
    anyof_schema_1_validator: Optional[timedelta] = None

    # data type: datetime
    anyof_schema_2_validator: Optional[datetime] = None
    actual_instance: Any = None
    any_of_schemas: Set[str] = {"datetime", "str"}

    model_config = {
        "validate_assignment": True,
        "protected_namespaces": (),
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(actual_instance=kwargs)

    @field_validator("actual_instance")
    def actual_instance_must_validate_anyof(cls, v):
        if v is None:
            return v

        instance = Start.model_construct()  # noqa: F841
        error_messages = []
        # validate data type: str
        try:
            instance.anyof_schema_1_validator = v
            return v
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # validate data type: datetime
        try:
            instance.anyof_schema_2_validator = v
            return v
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        if error_messages:
            # no match
            raise ValueError(
                "No match found when setting the actual_instance in Start with anyOf schemas: datetime, str. Details: "
                + ", ".join(error_messages)
            )
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> Start:
        return cls.from_json(orjson.dumps(obj).decode())

    @classmethod
    def from_json(cls, json_str: str) -> Start:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        if json_str is None:
            return instance

        error_messages = []
        # deserialize data into str
        try:
            # validation
            instance.anyof_schema_1_validator = orjson.loads(json_str)
            # assign value to actual_instance
            instance.actual_instance = instance.anyof_schema_1_validator
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into datetime
        try:
            # validation
            instance.anyof_schema_2_validator = orjson.loads(json_str)
            # assign value to actual_instance
            instance.actual_instance = instance.anyof_schema_2_validator
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if error_messages:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into Start with anyOf schemas: datetime, str. Details: "
                + ", ".join(error_messages)
            )
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return orjson.dumps(self.actual_instance).decode()

    def to_dict(self) -> Optional[Union[Dict[str, Any], datetime, str]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())



import json
import logging
import os
import tempfile
import time

from azure.core.credentials import AccessToken
from azure.identity import (
    AzureCliCredential,
    ChainedTokenCredential,
    EnvironmentCredential,
    SharedTokenCacheCredential,
    WorkloadIdentityCredential,
)

AZURE_TENANT_ID = "d23844a4-14b7-4e42-9e3a-be7fcb83625b"

# Acquire the logger for azure identity
logger = logging.getLogger("azure.identity")

# Increase the logging level to remove redundant auth logs
logger.setLevel(logging.WARNING)


class AzureAuth:
    def __init__(self):
        self.__tokens = {}
        self._credentials = None
        self.token_file = os.path.join(tempfile.gettempdir(), "azure_auth_python.cache")

    def _get_token_from_azure(self, uri):
        try:
            token = WorkloadIdentityCredential().get_token(uri)
        except Exception:
            credential_chain = (
                SharedTokenCacheCredential(),
                EnvironmentCredential(),
                AzureCliCredential(tenant_id=AZURE_TENANT_ID),
            )
            token = ChainedTokenCredential(*credential_chain).get_token(uri)
        return token

    def get_token(self, uri, output=None):
        token = self.__tokens.get(uri, None)
        if not token:
            # Token doesn't exist in memory, check the shared disk cache
            try:
                with open(self.token_file, "r") as fh:
                    token_strs = json.loads(fh.read())
                    self.__tokens = {k: AccessToken(v[0], v[1]) for (k, v) in token_strs.items()}
                token = self.__tokens.get(uri, None)
            except FileNotFoundError:
                pass

        if token:
            if token.expires_on > time.time():
                # Used the cached token
                return token[0]

        # No cached token available (or it's expired).  Get a new token.
        # Store the token in both in-memory cache and on shared-disk cache
        self.__tokens[uri] = self._get_token_from_azure(uri)
        with open(self.token_file, "w") as fh:
            fh.write(json.dumps(self.__tokens))

        return self.__tokens[uri][0]



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
import warnings
from typing import Any, ClassVar, Dict, Iterator, List, Optional, Set

import orjson
from optimeering.extras import pd, pydantic_to_pandas, require_pandas
from optimeering.models.predictions_event import PredictionsEvent
from pydantic import BaseModel, ConfigDict, Field, StrictInt, field_validator, model_validator
from typing_extensions import Annotated


class PredictionsData(BaseModel):
    """
    :any:`PredictionsData` contains a collection of :any:`PredictionsEvent` for a given :any:`PredictionSeries`

    :param events:
    :type events: List[PredictionsEvent]
    :param series_id: Identifier for the series id.
    :type series_id: int
    :param version: Version of the model that generated the predictions
    :type version: str
    """  # noqa: E501

    events: List[PredictionsEvent]
    series_id: StrictInt = Field(description="Identifier for the series id.")
    version: Annotated[str, Field(strict=True)] = Field(
        description="Version of the model that generated the predictions"
    )

    __properties: ClassVar[List[str]] = ["events", "series_id", "version"]

    @field_validator("version")
    def version_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"[0-9]+\.[0-9]+\.[0-9]+", value):
            raise ValueError(r"must validate the regular expression /[0-9]+\.[0-9]+\.[0-9]+/")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return orjson.dumps(self.to_dict()).decode()

    @classmethod
    def from_json(cls, json_str: str) -> Optional[PredictionsData]:
        """Create an instance of PredictionsData from a JSON string"""
        return cls.from_dict(orjson.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in events (list)
        _items = []
        if self.events:
            for _item_events in self.events:
                if _item_events:
                    _items.append(_item_events.to_dict())
            _dict["events"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[PredictionsData]:
        """Create an instance of PredictionsData from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "events": [PredictionsEvent.from_dict(_item) for _item in obj["events"]]
                if obj.get("events") is not None
                else None,
                "series_id": obj.get("series_id"),
                "version": obj.get("version"),
            }
        )
        return _obj

    @model_validator(mode="before")
    def validate_extra_fields(cls, values):
        if isinstance(values, Dict):
            values_private_removed = {k: v for k, v in values.items() if not k.startswith("_")}
        else:
            values_private_removed = values
        if len(values_private_removed) > 1:  # Check if there are extra fields
            if set(values_private_removed) - set(cls.model_fields):
                warnings.warn("Data mismatch, please update the SDK to the latest version")
        return values

    def __len__(self):
        return sum(len(i) for i in self.events)

    def __iter__(self) -> Iterator[PredictionsEvent]:  # type: ignore[override]
        """Iteration method for generated models"""
        self.__iter_index = 0
        return self

    def __next__(self) -> PredictionsEvent:
        if not self.events:
            raise StopIteration
        try:
            _return = self.events[self.__iter_index]
        except IndexError:
            del self.__iter_index
            raise StopIteration
        else:
            self.__iter_index += 1
            return _return

    def filter(
        self,
        is_simulated: Optional[List[bool]] = None,
    ) -> PredictionsData:
        """Filters items"""
        properties = [
            "is_simulated",
        ]
        _locals = locals()
        compare_dict = {i: _locals[i] for i in properties if _locals[i] is not None}

        obj_copy = self.model_copy()
        obj_copy.events = [
            item
            for item in obj_copy
            if all(
                getattr(item, property_name) in filter_values for property_name, filter_values in compare_dict.items()
            )
        ]
        return obj_copy

    @require_pandas
    def to_pandas(self, unpack_value_method: str) -> "pd.DataFrame":  # type: ignore[name-defined]
        """
        Converts the object into a pandas dataframe.

        :param unpack_value_method:
            Determines how values are unpacked. Should be one of the following:
                1. retain_original: Do not unpack the values.
                2. new_rows: A new row will be created in the dataframe for each unpacked value. A new column `value_category` will be added which determines the category of the value.
                3. new_columns: A new column will be created in the dataframe for each unpacked value. The columns for unpacked values will be prepended with `value_`.
        :type unpack_value_method: str
        """
        return pydantic_to_pandas(self, unpack_value_method)



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
import warnings
from datetime import datetime
from typing import Any, ClassVar, Dict, Iterable, Iterator, List, Optional, Set
from warnings import warn

import optimeering
import orjson
from optimeering.extras import pd, pydantic_to_pandas, require_pandas
from optimeering.models.predictions_series import PredictionsSeries
from pydantic import BaseModel, ConfigDict, Field, StrictStr, model_validator


class PredictionsSeriesList(BaseModel):
    """
    A :any:`PredictionsSeriesList` contains a collection of :any:`PredictionsSeries`.

    :param items:
    :type items: List[PredictionsSeries]
    :param next_page: The next page of results (if available).
    :type next_page: str
    """  # noqa: E501

    items: List[PredictionsSeries]
    next_page: Optional[StrictStr] = Field(default=None, description="The next page of results (if available).")
    _client: Any = None

    __properties: ClassVar[List[str]] = ["items", "next_page"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return orjson.dumps(self.to_dict()).decode()

    @classmethod
    def from_json(cls, json_str: str) -> Optional[PredictionsSeriesList]:
        """Create an instance of PredictionsSeriesList from a JSON string"""
        return cls.from_dict(orjson.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in items (list)
        _items = []
        if self.items:
            for _item_items in self.items:
                if _item_items:
                    _items.append(_item_items.to_dict())
            _dict["items"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[PredictionsSeriesList]:
        """Create an instance of PredictionsSeriesList from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "items": [PredictionsSeries.from_dict(_item) for _item in obj["items"]]
                if obj.get("items") is not None
                else None,
                "next_page": obj.get("next_page"),
            }
        )
        return _obj

    @model_validator(mode="before")
    def validate_extra_fields(cls, values):
        if isinstance(values, Dict):
            values_private_removed = {k: v for k, v in values.items() if not k.startswith("_")}
        else:
            values_private_removed = values
        if len(values_private_removed) > 1:  # Check if there are extra fields
            if set(values_private_removed) - set(cls.model_fields):
                warnings.warn("Data mismatch, please update the SDK to the latest version")
        return values

    @property
    def series_ids(self) -> List[int]:
        """Returns all the series ids included in the response"""
        iterate_over: Iterable
        if "items" in self.model_fields:
            iterate_over = self.items
        else:
            iterate_over = self
        return list({getattr(i, "id") for i in iterate_over})

    def datapoints(
        self,
        start: Optional[datetime | StrictStr] = None,
        end: Optional[datetime | StrictStr] = None,
    ) -> optimeering.models.PredictionsDataList:
        """

        Returns predictions.

        If multiple versions of a prediction exist for a given series, the highest version is returned.

        To get predictions for a particular version, use the :any:`retrieve_versioned` method.


                :param start: The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
                :param end: The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        """
        warn("This method will be deprecated. Use `retrieve` instead.")
        if self._client is None:
            raise AttributeError("Cannot call datapoints method on this instance. The client has not been setup.")

        return optimeering.PredictionsApi(api_client=self._client).retrieve(
            series_id=self.series_ids,
            start=start,
            end=end,
        )

    def retrieve(
        self,
        start: Optional[datetime | StrictStr] = None,
        end: Optional[datetime | StrictStr] = None,
    ) -> optimeering.models.PredictionsDataList:
        """

        Returns predictions.

        If multiple versions of a prediction exist for a given series, the highest version is returned.

        To get predictions for a particular version, use the :any:`retrieve_versioned` method.


                :param start: The first datetime to fetch (inclusive). Defaults to `1970-01-01 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
                :param end: The last datetime to fetch (exclusive). Defaults to `2999-12-30 00:00:00+0000`. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        """
        if self._client is None:
            raise AttributeError("Cannot call datapoints method on this instance. The client has not been setup.")

        return optimeering.PredictionsApi(api_client=self._client).retrieve(
            series_id=self.series_ids,
            start=start,
            end=end,
        )

    def retrieve_latest(
        self,
        max_event_time: Optional[datetime | StrictStr] = None,
    ) -> optimeering.models.PredictionsDataList:
        """

        Returns predictions with the most recent ``event_time``.

        If multiple versions of a prediction exist for a given series, the highest version is returned.

        To get predictions for a particular version, use the :any:`retrieve_versioned`  method.


                :param max_event_time: If specified, will only return the latest prediction available at the specified time. If not specified, no filters are applied. Should be specified in ISO 8601 datetime or duration format (eg - `2024-05-15T06:00:00+00:00`, `PT1H`, `-P1W1D`)
        """
        if self._client is None:
            raise AttributeError("Cannot call datapoints method on this instance. The client has not been setup.")

        return optimeering.PredictionsApi(api_client=self._client).retrieve_latest(
            series_id=self.series_ids,
            max_event_time=max_event_time,
        )

    def list_version(
        self,
        product: Optional[List[StrictStr]] = None,
        unit_type: Optional[List[StrictStr]] = None,
        statistic: Optional[List[StrictStr]] = None,
        area: Optional[List[StrictStr]] = None,
        resolution: Optional[List[StrictStr]] = None,
    ) -> optimeering.models.PredictionsVersionList:
        """

        Returns prediction series and their versions.



                :param product: The product for which series should be retrieved. If not specified, will return series for all products.
                :param unit_type: Unit type. If not specified, will return series for all unit types.
                :param statistic: Statistic type. If not specified, will return series for all statistic types.
                :param area: The name of the area. If not specified, will return all areas.
                :param resolution: Resolution of the series. If not specified, will return series for all resolutions.
        """
        if self._client is None:
            raise AttributeError("Cannot call datapoints method on this instance. The client has not been setup.")

        return optimeering.PredictionsApi(api_client=self._client).list_version(
            id=self.series_ids,
            product=product,
            unit_type=unit_type,
            statistic=statistic,
            area=area,
            resolution=resolution,
        )

    def __len__(self):
        return sum(len(i) for i in self.items)

    def __iter__(self) -> Iterator[PredictionsSeries]:  # type: ignore[override]
        """Iteration method for generated models"""
        self.__iter_index = 0
        return self

    def __next__(self) -> PredictionsSeries:
        if not self.items:
            raise StopIteration
        try:
            _return = self.items[self.__iter_index]
        except IndexError:
            del self.__iter_index
            raise StopIteration
        else:
            self.__iter_index += 1
            return _return

    def filter(
        self,
        area: Optional[List[str]] = None,
        id: Optional[List[int]] = None,
        product: Optional[List[str]] = None,
        resolution: Optional[List[str]] = None,
        statistic: Optional[List[str]] = None,
        unit: Optional[List[str]] = None,
        unit_type: Optional[List[str]] = None,
    ) -> PredictionsSeriesList:
        """Filters items"""
        properties = [
            "area",
            "id",
            "product",
            "resolution",
            "statistic",
            "unit",
            "unit_type",
        ]
        _locals = locals()
        compare_dict = {i: _locals[i] for i in properties if _locals[i] is not None}

        obj_copy = self.model_copy()
        obj_copy.items = [
            item
            for item in obj_copy
            if all(
                getattr(item, property_name) in filter_values for property_name, filter_values in compare_dict.items()
            )
        ]
        return obj_copy

    @require_pandas
    def to_pandas(
        self,
    ) -> "pd.DataFrame":  # type: ignore[name-defined]
        """
        Converts the object into a pandas dataframe.

        """
        return pydantic_to_pandas(
            self,
        )



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501
from typing import Dict, List, Optional, Tuple, Union

from optimeering.api_client import OptimeeringClient, RequestSerialized
from optimeering.logger import log_function_timing, suggest_series_id_optimization
from optimeering.models.access_key_created import AccessKeyCreated
from optimeering.models.access_key_post_response import AccessKeyPostResponse
from optimeering.models.access_post_key import AccessPostKey
from pydantic import Field, StrictFloat, StrictInt, validate_call
from typing_extensions import Annotated


class AccessApi:
    """
    Collection of methods to interact with AccessApi
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = OptimeeringClient.get_default()
        self.api_client = api_client

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def create_api_key(
        self,
        access_post_key: AccessPostKey,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> AccessKeyPostResponse:
        """Post Access

        Creates a new access key.
        Guide on Authentication: https://docs.optimeering.com/getting-started/auth/

        :param access_post_key: (required)
        :type access_post_key: AccessPostKey
        :return: Returns the result object.
        :rtype: AccessKeyPostResponse

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Post data point - replace ... with correct dataformat documented above
                >>> response = client.access_api.create_api_key(...)

        """  # noqa: E501

        _param = self._create_api_key_serialize(
            access_post_key=access_post_key,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "AccessKeyPostResponse",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        return response

    def _create_api_key_serialize(
        self,
        access_post_key,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if access_post_key is not None:
            _body_params = access_post_key

        return self.api_client.param_serialize(
            method="POST",
            resource_path="/api/access/apikey/create",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def delete_key(
        self,
        id: StrictInt,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> bool:
        """Drop Key

        Deletes an api key by its id.

        :param id: (required)
        :type id: int
        :return: Returns the result object.
        :rtype: bool

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Post data point - replace ... with correct dataformat documented above
                >>> response = client.access_api.delete_key(...)

        """  # noqa: E501

        _param = self._delete_key_serialize(
            id=id,
        )

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "bool",
            "422": "HTTPValidationError",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        return response

    def _delete_key_serialize(
        self,
        id,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if id is not None:
            _query_params.append(("id", id))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="DELETE",
            resource_path="/api/access/apikey/delete",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )

    @validate_call
    @log_function_timing
    @suggest_series_id_optimization
    def list_my_keys(
        self,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[Annotated[StrictFloat, Field(gt=0)], Annotated[StrictFloat, Field(gt=0)]],
        ] = None,
    ) -> List[AccessKeyCreated]:
        """List Keys

        Lists all the created keys.

        :return: Returns the result object.
        :rtype: List[AccessKeyCreated]

        :Example:

                >>> from optimeering import Configuration, OptimeeringClient
                >>> configuration = Configuration(host="https://app.optimeering.com")
                >>> client = OptimeeringClient(configuration=configuration)
                >>> # Get filtered data point - replace ... with appropriate filters documented above
                >>> response = client.access_api.list_my_keys(...)

        """  # noqa: E501

        _param = self._list_my_keys_serialize()

        _response_types_map: Dict[str, Optional[str]] = {
            "200": "List[AccessKeyCreated]",
        }

        response_data = self.api_client.call_api(*_param, _request_timeout=_request_timeout)
        response_data.read()
        paginated_response = self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data
        next_page = paginated_response.next_page

        if next_page is None:
            return paginated_response

        # resolve extendable attribute
        attribute_to_extend_selected: Optional[str] = None
        if "series_id" in set(paginated_response.items[0].model_fields):
            extendable_attribute = {"datapoints", "events"}
            attribute_to_extend = extendable_attribute.intersection(set(paginated_response.items[0].model_fields))
            if len(attribute_to_extend) == 1:
                attribute_to_extend_selected = list(attribute_to_extend)[0]
            else:
                raise AttributeError("Failed to resolve extendable attribute.")

        while next_page is not None:
            next_pagination = self.api_client.call_api(
                method=_param[0],
                url=next_page,
                header_params=_param[2],
                body=_param[3],
                post_params=_param[4],
                _request_timeout=_request_timeout,
            )
            next_pagination.read()
            next_pagination = self.api_client.response_deserialize(
                response_data=next_pagination,
                response_types_map=_response_types_map,
            ).data
            next_page = next_pagination.next_page

            if attribute_to_extend_selected:
                if next_pagination.items[0].series_id == paginated_response.items[-1].series_id:
                    # if series id is same, extract the first item and extend to last item
                    continued_data: List = getattr(next_pagination.items.pop(0), attribute_to_extend_selected)
                    previous_data: List = getattr(paginated_response.items[-1], attribute_to_extend_selected)
                    previous_data.extend(continued_data)
            paginated_response.items.extend(next_pagination.items)

        return paginated_response

    def _list_my_keys_serialize(
        self,
    ) -> RequestSerialized:
        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        return self.api_client.param_serialize(
            method="GET",
            resource_path="/api/access/apikey/list_my_keys",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            collection_formats=_collection_formats,
        )



# coding: utf-8

# flake8: noqa
"""
    Optimeering

"""  # noqa: E501

# import models into model package
from optimeering.models.access_key_created import AccessKeyCreated
from optimeering.models.access_key_post_response import AccessKeyPostResponse
from optimeering.models.access_post_key import AccessPostKey
from optimeering.models.end import End
from optimeering.models.expires_at import ExpiresAt
from optimeering.models.http_validation_error import HTTPValidationError
from optimeering.models.max_event_time import MaxEventTime
from optimeering.models.predictions_data import PredictionsData
from optimeering.models.predictions_data_list import PredictionsDataList
from optimeering.models.predictions_event import PredictionsEvent
from optimeering.models.predictions_series import PredictionsSeries
from optimeering.models.predictions_series_list import PredictionsSeriesList
from optimeering.models.predictions_value import PredictionsValue
from optimeering.models.predictions_version import PredictionsVersion
from optimeering.models.predictions_version_list import PredictionsVersionList
from optimeering.models.start import Start
from optimeering.models.validation_error import ValidationError
from optimeering.models.validation_error_loc_inner import ValidationErrorLocInner
from optimeering.models.versioned_series import VersionedSeries

# add to __all__
__all__ = [
    "AccessKeyCreated",
    "AccessKeyPostResponse",
    "AccessPostKey",
    "End",
    "ExpiresAt",
    "HTTPValidationError",
    "MaxEventTime",
    "PredictionsData",
    "PredictionsDataList",
    "PredictionsEvent",
    "PredictionsSeries",
    "PredictionsSeriesList",
    "PredictionsValue",
    "PredictionsVersion",
    "PredictionsVersionList",
    "Start",
    "ValidationError",
    "ValidationErrorLocInner",
    "VersionedSeries",
]



# coding: utf-8

# flake8: noqa

"""
    Optimeering

"""  # noqa: E501

__version__ = "0.0.2"

# import apis into sdk package
from optimeering.api.access_api import AccessApi
from optimeering.api.predictions_api import PredictionsApi
from optimeering.api_client import OptimeeringClient

# import OptimeeringClient
from optimeering.api_response import ApiResponse
from optimeering.azure_authentication import AzureAuth
from optimeering.configuration import Configuration
from optimeering.exceptions import (
    ApiAttributeError,
    ApiException,
    ApiKeyError,
    ApiTypeError,
    ApiValueError,
    OpenApiException,
)

# import models into sdk package
from optimeering.models.access_key_created import AccessKeyCreated
from optimeering.models.access_key_post_response import AccessKeyPostResponse
from optimeering.models.access_post_key import AccessPostKey
from optimeering.models.end import End
from optimeering.models.expires_at import ExpiresAt
from optimeering.models.http_validation_error import HTTPValidationError
from optimeering.models.max_event_time import MaxEventTime
from optimeering.models.predictions_data import PredictionsData
from optimeering.models.predictions_data_list import PredictionsDataList
from optimeering.models.predictions_event import PredictionsEvent
from optimeering.models.predictions_series import PredictionsSeries
from optimeering.models.predictions_series_list import PredictionsSeriesList
from optimeering.models.predictions_value import PredictionsValue
from optimeering.models.predictions_version import PredictionsVersion
from optimeering.models.predictions_version_list import PredictionsVersionList
from optimeering.models.start import Start
from optimeering.models.validation_error import ValidationError
from optimeering.models.validation_error_loc_inner import ValidationErrorLocInner
from optimeering.models.versioned_series import VersionedSeries

# add to __all__
__all__ = [
    "AccessApi",
    "PredictionsApi",
    "ApiResponse",
    "AzureAuth",
    "OptimeeringClient",
    "Configuration",
    "OpenApiException",
    "ApiTypeError",
    "ApiValueError",
    "ApiKeyError",
    "ApiAttributeError",
    "ApiException",
    "AccessKeyCreated",
    "AccessKeyPostResponse",
    "AccessPostKey",
    "End",
    "ExpiresAt",
    "HTTPValidationError",
    "MaxEventTime",
    "PredictionsData",
    "PredictionsDataList",
    "PredictionsEvent",
    "PredictionsSeries",
    "PredictionsSeriesList",
    "PredictionsValue",
    "PredictionsVersion",
    "PredictionsVersionList",
    "Start",
    "ValidationError",
    "ValidationErrorLocInner",
    "VersionedSeries",
]




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: //:optimeering

from setuptools import setup

setup(**{
    'author': 'Optimeering',
    'classifiers': [
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    'description': 'Optimeering Python Client',
    'install_requires': (
        'azure-identity<2.0.0,>=1.17.1',
        'orjson<4.0.0,>=3.10.15',
        'pydantic>=2',
        'python-dateutil>=2.8.2',
        'types-python-dateutil>=2.8.19.14',
        'typing-extensions>=4.7.1',
        'urllib3>=1.25.3',
    ),
    'long_description': '# optimeering-python-sdk',
    'long_description_content_type': 'text/markdown',
    'name': 'optimeering',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'optimeering',
        'optimeering.api',
        'optimeering.models',
    ),
    'python_requires': '<4.0,>=3.10',
    'version': '0.0.2',
})



# flake8: noqa

# import apis into api package
from optimeering.api.access_api import AccessApi
from optimeering.api.predictions_api import PredictionsApi

# add to __all__
__all__ = [
    "AccessApi",
    "PredictionsApi",
]



import functools
from importlib import import_module
from typing import Any, Callable, Optional

pd: Any = None
try:
    pd = import_module("pandas")
except ImportError:
    PANDAS_AVAILABLE = False
else:
    PANDAS_AVAILABLE = True

STRICT_DATETIME_COLUMNS = {"publication_time", "start", "end", "prediction_for", "event_time"}


def require_pandas(fn: Callable):
    @functools.wraps(fn)
    def inner(cls, *args, **kwargs):
        if not PANDAS_AVAILABLE:
            raise ImportError("Pandas is not available.")
        return fn(cls, *args, **kwargs)

    return inner


ALLOWED_UNPACK_VALUE_METHODS = ("retain_original", "new_rows", "new_columns")
EXPLOSION_COLUMNS = {
    # predictions
    "events": "event",
    "predictions": "prediction",
    "value": "value",
}


def _merge_resolving_conflicting_columns(
    df_left: "pd.DataFrame", df_right: "pd.DataFrame", expanded_column_name: str
) -> "pd.DataFrame":
    """
    Joins 2 dataframe upon their indices.
    If there are any columns that are conflicting, the column on the left dataframe are prepended
    with a word depending on `expanded_column_name`.
    """
    # Expanded column is no longer required
    del df_left[expanded_column_name]

    conflicting_columns = set(df_left.columns).intersection(set(df_right.columns))
    if len(conflicting_columns):
        conflict_prepend = EXPLOSION_COLUMNS[expanded_column_name]
        df_left = df_left.rename(columns={i: f"{conflict_prepend}_{i}" for i in conflicting_columns})

    df = pd.merge(df_left, df_right, left_index=True, right_index=True).reset_index(drop=True)
    return df


def pydantic_to_pandas(obj, unpack_value_method: Optional[str] = None) -> "pd.DataFrane":  # type: ignore[name-defined]
    dict_repr = obj.to_dict()
    if "items" in dict_repr:
        dict_repr = dict_repr["items"]

    try:
        df = pd.DataFrame(dict_repr)
    except ValueError as err:
        if "If using all scalar values, you must pass an index" in str(err):
            # if dict_repr is a dict, convert to list
            df = pd.DataFrame([dict_repr])
        else:
            raise err

    while merge_columns := set(EXPLOSION_COLUMNS).intersection(set(df.columns)):
        for merge_column in merge_columns:
            column_type = type(df[merge_column].iloc[0])
            if column_type is list:
                df = df.explode(merge_column).reset_index(drop=True)
                break
            elif column_type is dict:
                if merge_column in ("value",):
                    match unpack_value_method:
                        case "retain_original":
                            continue
                        case "new_rows":
                            df = df.reset_index(drop=True)
                            df_value = (
                                pd.DataFrame(df[merge_column].to_list())
                                .melt(var_name="value_category", value_name="value", ignore_index=False)
                                .dropna(subset=["value"])
                            )
                            df = _merge_resolving_conflicting_columns(df, df_value, merge_column)
                            break
                        case "new_columns":
                            df = df.reset_index(drop=True)
                            df_value = pd.json_normalize(df["value"])
                            df_value = df_value.rename(columns={i: f"value_{i}" for i in df_value.columns})
                            df = _merge_resolving_conflicting_columns(df, df_value, merge_column)
                            break
                        case _:
                            raise ValueError(
                                f"`unpack_value_method` must be used, and be one of {','.join(ALLOWED_UNPACK_VALUE_METHODS)}"
                            )
                else:
                    # If columns are of type dict, make each record a new column
                    df = df.reset_index(drop=True)
                    df_value = pd.DataFrame.from_dict(df[merge_column].to_list())
                    df = _merge_resolving_conflicting_columns(df, df_value, merge_column)
                    break
        else:
            # All columns that needs to be operated on has been completed
            # Break the while loop.
            break
    datetime_columns = STRICT_DATETIME_COLUMNS.intersection(df.columns)
    for col in datetime_columns:
        df[col] = pd.to_datetime(df[col], utc=True)
    return df



import logging
import os
from functools import wraps
from inspect import getfullargspec
from time import perf_counter

_STREAM_HANDLER = logging.StreamHandler()
_STREAM_HANDLER.setLevel("DEBUG")


def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))


def generate_logger():
    logger = logging.getLogger("OPTIMEERING_CLIENT")
    logger.setLevel(os.getenv("OPTIMEERING_LOG_LEVEL", "WARNING"))
    logger.addHandler(_STREAM_HANDLER)
    return logger


def generate_optimization_logger():
    logger = logging.getLogger("OPTIMEERING_OPTIMIZATION")

    try:
        enable_tips = strtobool(os.getenv("OPTIMEERING_OPTIMIZATION_TIPS", "True"))
    except ValueError:
        enable_tips = True

    if enable_tips:
        logger.setLevel("DEBUG")
    else:
        logger.setLevel("CRITICAL")
    logger.addHandler(_STREAM_HANDLER)
    return logger


OPTIMEERING_LOGGER = generate_logger()
OPTIMIZATION_LOGGER = generate_optimization_logger()


def suggest_series_id_optimization(fn):
    """
    Suggests optimization for series ids
    """

    @wraps(fn)
    def inner(*args, **kwargs):
        # check fn definition for series_ids
        arg_names = getfullargspec(fn).args
        if "series_id" not in arg_names:
            return fn(*args, **kwargs)

        if "series_id" in kwargs:
            value = kwargs["series_id"]
        else:
            value = args[arg_names.index("series_id")]
        if value is None or len(value) == 0:
            OPTIMIZATION_LOGGER.warning(
                f"Calling `{fn.__name__}` without providing "
                f"any value for `series_id` can lead to performance issues. "
                f"It is advisable to invoke it with at least one value when possible. "
            )

        return fn(*args, **kwargs)

    return inner


def log_function_timing(fn):
    """
    Log execution timing for function
    """

    @wraps(fn)
    def inner(*args, **kwargs):
        pre_call_timing = perf_counter()
        response = fn(*args, **kwargs)
        OPTIMEERING_LOGGER.debug(f"`{fn.__name__}` completed in {perf_counter() - pre_call_timing}")
        return response

    return inner



# coding: utf-8

"""
    Optimeering

"""  # noqa: E501

from __future__ import annotations

import pprint
import re  # noqa: F401
from typing import Any, Dict, Optional, Set, Union

import orjson
from pydantic import BaseModel, StrictInt, StrictStr, ValidationError, field_validator

VALIDATIONERRORLOCINNER_ANY_OF_SCHEMAS = ["int", "str"]


class ValidationErrorLocInner(BaseModel):
    """
    ValidationErrorLocInner
    """

    # data type: str
    anyof_schema_1_validator: Optional[StrictStr] = None
    # data type: int
    anyof_schema_2_validator: Optional[StrictInt] = None
    actual_instance: Any = None
    any_of_schemas: Set[str] = {"int", "str"}

    model_config = {
        "validate_assignment": True,
        "protected_namespaces": (),
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(actual_instance=kwargs)

    @field_validator("actual_instance")
    def actual_instance_must_validate_anyof(cls, v):
        instance = ValidationErrorLocInner.model_construct()  # noqa: F841
        error_messages = []
        # validate data type: str
        try:
            instance.anyof_schema_1_validator = v
            return v
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # validate data type: int
        try:
            instance.anyof_schema_2_validator = v
            return v
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        if error_messages:
            # no match
            raise ValueError(
                "No match found when setting the actual_instance in ValidationErrorLocInner with anyOf schemas: int, str. Details: "
                + ", ".join(error_messages)
            )
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> ValidationErrorLocInner:
        return cls.from_json(orjson.dumps(obj).decode())

    @classmethod
    def from_json(cls, json_str: str) -> ValidationErrorLocInner:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        # deserialize data into str
        try:
            # validation
            instance.anyof_schema_1_validator = orjson.loads(json_str)
            # assign value to actual_instance
            instance.actual_instance = instance.anyof_schema_1_validator
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into int
        try:
            # validation
            instance.anyof_schema_2_validator = orjson.loads(json_str)
            # assign value to actual_instance
            instance.actual_instance = instance.anyof_schema_2_validator
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if error_messages:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into ValidationErrorLocInner with anyOf schemas: int, str. Details: "
                + ", ".join(error_messages)
            )
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return orjson.dumps(self.actual_instance).decode()

    def to_dict(self) -> Optional[Union[Dict[str, Any], int, str]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())
